GLOBAL.setmetatable(env,{__index=function(t,k) return GLOBAL.rawget(GLOBAL,k) end})

-- Only usefful in hamlet!
if not IsDLCEnabled(3) then return end

local function IsInBounds(inst, pt)
    local interiorSpawner = GetWorld().components.interiorspawner

    if interiorSpawner.current_interior then
        local width = interiorSpawner.current_interior.width
        local depth = interiorSpawner.current_interior.depth
        local originpt = interiorSpawner:getSpawnOrigin()

        local dMax = originpt.x + depth/2
        local dMin = originpt.x - depth/2

        local wMax = originpt.z + width/2
        local wMin = originpt.z - width/2

        local dist = 0.2

        if pt.x < dMin+dist or pt.x > dMax -dist or pt.z < wMin+dist or pt.z > wMax-dist then
            return false
        end
    end

    return true
end

local function PlaceTestRugFn(inst, pt)
    if inst.parent then
        local myrot = inst.Transform:GetRotation()
        local px, py, pz = inst.Transform:GetWorldPosition()
        inst.parent:RemoveChild(inst)
        inst.Transform:SetPosition(px,py,pz)
        inst.Transform:SetRotation(myrot)
    end

    return IsInBounds(inst, pt)
end

local function placeTestCeilingLightFn(inst, pt)
    if inst.parent then
        local px, py, pz = inst.Transform:GetWorldPosition()
        inst.parent:RemoveChild(inst)
        inst.Transform:SetPosition(px,py,pz)
    end

    inst.AnimState:SetOrientation(ANIM_ORIENTATION.RotatingBillboard)  
    inst.Transform:SetRotation(-90)

    return IsInBounds(inst, pt)
end

local function NonWalledShelves(inst)
    inst.AnimState:SetLayer(LAYER_WORLD)
    inst.AnimState:SetSortOrder(0)

    if table.contains({"shelves_pipe", "shelves_hattree"}, inst.prefab) then
        -- Rounded Physics.
        inst.Physics:SetCapsule(0.5, 2)
    end
end

--------------------------------------------------------------------------------------------------

local function ChangeRecipes(recipe)
    if not (recipe.level.HOME == 2 and recipe.placer) then
        return
    end

    -- Detachs the shelves from the wall.
    if recipe.tab == RENO_RECIPETABS.SHELVES then
        AddPrefabPostInit(recipe.name, NonWalledShelves)
    end

    -- Compability with "Decor Galore" Mod.
    if table.contains({ RENO_RECIPETABS.SHELVES, RENO_RECIPETABS.MISC }, recipe.tab) then
        recipe.wallitem = nil
    end

    local NewPlaceTestFn = 
        (recipe.tab == RENO_RECIPETABS.RUGS          and PlaceTestRugFn         ) or
        (recipe.tab == RENO_RECIPETABS.HANGING_LAMPS and placeTestCeilingLightFn) or
        IsInBounds

    if not recipe.wallitem then
        -- Forniture can be placed without minimun spacing!
        AddPrefabPostInit(recipe.placer, function(inst)
            inst.components.placer.placeTestFn = NewPlaceTestFn
        end)
    end
end

for _, recipe in pairs(GetAllRecipes(true)) do
    ChangeRecipes(recipe)
end

--------------------------------------------------------------------------------------------------

-- Things can be closer in interiors.
local function CanBuildWithLessMinSpace(self, pt, recipe)
    if TheCamera.interior then
        if recipe.decor then
            return IsInBounds(nil, pt)
        end

        local tile = GetWorld().Map:GetTileAtPoint(pt:Get())
        if tile == GROUND.IMPASSABLE or recipe.aquatic then
            return false
        end

        local ents = TheSim:FindEntities(pt.x, pt.y, pt.z, 6, nil, {'player', 'fx', 'NOBLOCK', "isinventoryitem"})

        for k, v in pairs(ents) do
            if v ~= self.inst and (not v.components.placer) and v.entity:IsVisible() and not (v.components.inventoryitem and v.components.inventoryitem.owner) then
                local dsq = distsq(_G.Vector3(v.Transform:GetWorldPosition()), pt)

                return dsq > 1
            end
        end

        return true
    end

    return false
end

AddComponentPostInit("builder", function(self)
    local _CanBuildAtPoint = self.CanBuildAtPoint

    function self:CanBuildAtPoint(pt, recipe)
        return _CanBuildAtPoint(self, pt, recipe) or CanBuildWithLessMinSpace(self, pt, recipe)
    end
end)